"""
Usage:
python3 make_delta --base ~/model_weights/llama-7b --target ~/model_weights/shikra-7b --delta ~/model_weights/shikra-7b-delta
"""
import argparse

import torch
from tqdm import tqdm
from transformers import AutoTokenizer, AutoModelForCausalLM
from shikra import ShikraLlamaForCausalLM


def make_delta(base_model_path, target_model_path, delta_path, hub_repo_id):
    print("Loading base model")
    base = AutoModelForCausalLM.from_pretrained(base_model_path, torch_dtype=torch.float16, low_cpu_mem_usage=True)

    print("Loading target model")
    target = ShikraLlamaForCausalLM.from_pretrained(target_model_path, torch_dtype=torch.float16, low_cpu_mem_usage=True)

    print("Calculating delta")
    for name, param in tqdm(target._dict().items(), desc="Calculating delta"):
        if name not in base._dict():
            assert name in ['model.mm_projector.weight', 'model.mm_projector.bias'], f'{name} not in base model'
            continue
        if param.data.shape == base._dict()[name].shape:
            param.data -= base._dict()[name]
        else:
            assert name in ['model.embed_tokens.weight',
                            'lm_head.weight'], f'{name} dimension mismatch: {param.data.shape} vs {base._dict()[name].shape}'
            bparam = base._dict()[name]
            param.data[:bparam.shape[0], :bparam.shape[1]] -= bparam

    print("Saving delta")
    if hub_repo_id:
        kwargs = {"push_to_hub": True, "repo_id": hub_repo_id}
    else:
        kwargs = {}
    target.save_pretrained(delta_path, **kwargs)
    target_tokenizer = AutoTokenizer.from_pretrained(target_model_path)
    target_tokenizer.save_pretrained(delta_path, **kwargs)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--base", type=str, required=True)
    parser.add_argument("--target", type=str, required=True)
    parser.add_argument("--delta", type=str, required=True)
    args = parser.parse_args()

    make_delta(args.base, args.target, args.delta, None)
